import sys
sys.path.append("./")

from morphapi.utils.data_io import connected_to_internet

import requests
import time


mouselight_base_url = "http://ml-neuronbrowser.janelia.org/"


def request(url):
	"""
	Sends a request to a url

	:param url: 

	"""
	if not connected_to_internet():
		raise ConnectionError("You need to have an internet connection to send requests.")
	response = requests.get(url)
	if response.ok:
		return response
	else:
		exception_string = 'URL request failed: {}'.format(response.reason)
	raise ValueError(exception_string)

def query_mouselight(query):
	"""
	Sends a GET request, not currently used for anything.

	:param query: 

	"""
	if not connected_to_internet():
		raise ConnectionError("You need an internet connection for API queries, sorry.")
	
	base_url = "http://ml-neuronbrowser.janelia.org/"

	full_query = base_url + query

	# send the query, package the return argument as a json tree
	response = requests.get(full_query)
	if response.ok:
		json_tree = response.json()
		if json_tree['success']:
			return json_tree
		else:
			exception_string = 'did not complete api query successfully'
	else:
		exception_string = 'API failure. Allen says: {}'.format(
			response.reason)

	# raise an exception if the API request failed
	raise ValueError(exception_string)

def post_mouselight(url, query=None, clean=False, attempts=3):
	"""
	sends a POST request to a user URL. Query can be either a string (in which case clean should be False) or a dictionary.

	:param url: 
	:param query: string or dictionary with query   (Default value = None)
	:param clean: if not clean, the query is assumed to be in JSON format (Default value = False)
	:param attempts: number of attempts  (Default value = 3)

	"""
	if not connected_to_internet():
		raise ConnectionError("You need an internet connection for API queries, sorry.")

	request = None
	if query is not None:
		for i in range(attempts):
			try:
				if not clean:
					time.sleep(0.01) # avoid getting an error from server
					request = requests.post(url, json={'query': query})
				else:
					time.sleep(0.01) # avoid getting an error from server
					request = requests.post(url, json=query)
			except Exception as e:
				exception = e
				request = None
				print('MouseLight API query failed. Attempt {} of {}'.format(i+1, attempts))
			if request is not None: break

		if request is None:
			raise ConnectionError("\n\nMouseLight API query failed with error message:\n{}.\
						\nPerhaps the server is down, visit 'http://ml-neuronbrowser.janelia.org' to find out.".format(exception))
	else:
		raise  NotImplementedError
	
	if request.status_code == 200:
		jreq = request.json()
		if 'data' in list(jreq.keys()):
			return jreq['data']
		else:
			return jreq
	else:
		raise Exception("Query failed to run by returning code of {}. {} -- \n\n{}".format(request.status_code, query, request.text))



from vtkplotter import Plotter

from morphapi.api.mouselight import MouseLightAPI


# ---------------------------- Downloading neurons --------------------------- #
mlapi = MouseLightAPI()

# Fetch metadata for neurons with some in the secondary motor cortex
neurons_metadata = mlapi.fetch_neurons_metadata(filterby='soma', filter_regions=['MOs'])

neurons = mlapi.download_neurons(neurons_metadata[0]) # downloading only one neuron to speed things up

# ------------------------------- Visualisation ------------------------------ #
print('creating meshes')
neurons = [neuron.create_mesh()[1] for neuron in neurons]

print('visualizing')
vp = Plotter(shape=(1, len(neurons)), axes=1)

vp.show(neurons)


import os
from collections import namedtuple
import numpy as np

from vtkplotter.shapes import Sphere, Tube
from vtkplotter import merge, write
from vtkplotter.colors import colorMap

import neurom as nm
from neurom.core import iter_neurites, iter_segments, iter_sections
from neurom.core.dataformat import COLS
from neurom.morphmath import segment_radius
from neurom.view.view import _get_linewidth

component = namedtuple("component", "x y z coords radius component")

class Neuron:
    _neurite_types = {'basal_dendrites':nm.core.types.NeuriteType.basal_dendrite, 
                        'apical_dendrites':nm.core.types.NeuriteType.apical_dendrite, 
                        'axon':nm.core.types.NeuriteType.axon}

    _ntypes = nm.core.types.NEURITES

    def __init__(self, swc_file=None, json_file=None):
        # Parse agruments
        if swc_file is not None and json_file is not None:
            raise ValueError('Pass eithr swc or json file, not both')
        elif swc_file is not None:
            if not os.path.exists(swc_file) or not swc_file.endswith('.swc'):
                raise ValueError(f'Invalid swc file: {swc_file}')
            self.data_file = swc_file
            self.data_file_type = 'swc'
        elif json_file is not None:
            if not os.path.exists(json_file) or not json_file.endswith('.json'):
                raise ValueError(f'Invalid json file: {json_file}')
            self.data_file = json_file
            self.data_file_type = 'json'
        else:
            self.data_file = None
            self.data_file_type = None
            self.morphology = None
            self.points = None

        if self.data_file is not None:
            self.load_from_file()


    def _parse_mesh_kwargs(self, **kwargs):
        # To give the entire neuron the same color
        neuron_color = kwargs.pop('neuron_color', None)

        # To give the entire neuron a color based on a cmap 
        neuron_number = kwargs.pop('neuron_number', None)
        cmap_lims = kwargs.pop('cmap_lims', (-1, 1))
        cmap = kwargs.pop('cmap', None)

        # To color each component individually
        soma_color = kwargs.pop('soma_color', 'salmon')
        apical_dendrites_color = kwargs.pop('apical_dendrites_color', 'salmon')
        basal_dendrites_color = kwargs.pop('basal_dendrites_color', apical_dendrites_color) 
        axon_color = kwargs.pop('axon_color', 'salmon')
        whole_neuron_color = kwargs.pop('whole_neuron_color', None)

        # Get each components color from args
        if neuron_color is not None:  # uniform color
            soma_color = apical_dendrites_color = basal_dendrites_color = axon_color = neuron_color
        elif cmap is not None:  # color according to cmap
            if neuron_number is None:
                neuron_number = 0
            
            soma_color = colorMap(neuron_number, name=cmap, vmin=cmap_lims[0], vmax=cmap_lims[1])
            apical_dendrites_color = basal_dendrites_color = axon_color = soma_color

        else: # Use color specified for each component
            pass

        if whole_neuron_color is None: whole_neuron_color = soma_color
        return soma_color, apical_dendrites_color, basal_dendrites_color, axon_color, whole_neuron_color, kwargs


    def load_from_file(self):
        if self.data_file_type is None:
            return
        elif self.data_file_type == 'json':
            raise NotImplementedError
        else:
            self.load_from_swc()

    def load_from_swc(self):
        nrn = nm.load_neuron(self.data_file)

        # Get position and radius of some
        soma_pos = nrn.soma.points[0, :3]
        soma_radius = nrn.soma.points[0, -1]

        # Get the rest of the data and store it
        self.morphology = nrn
        self.points = dict(
            soma=component(soma_pos[0], soma_pos[1], soma_pos[2], soma_pos, soma_radius, nrn.soma),
            )

        for ntype, nclass in self._neurite_types.items():
            self.points[ntype] = [component(n.points[:, 0], n.points[:, 1], n.points[:, 2], n.points[:, :3], n.points[:, -1], n) 
                                        for n in nrn.neurites if n.type == nclass]


    def create_mesh(self, fixed_neurite_radius=False, **kwargs):
        if self.points is None:
            print('No data loaded, returning')
            return

        # Parse kwargs
        soma_color, apical_dendrites_color, basal_dendrites_color, \
                    axon_color, whole_neuron_color, kwargs = \
                                        self._parse_mesh_kwargs(**kwargs)

        if fixed_neurite_radius is None: fixed_neurite_radius = False
        if fixed_neurite_radius:
            if not isinstance(fixed_neurite_radius, (int, float)) or not fixed_neurite_radius > 0:
                raise ValueError(f'Invalid value for parameter fixed_neurite_radius, should be a float > 0')

        # Create soma actor
        neurites = {}
        soma = Sphere(pos=self.points['soma'].coords, r=self.points['soma'].radius, c=soma_color)
        neurites['soma'] = [soma.clone().c(soma_color)]

        # Create neurites actors
        for ntype in self._neurite_types:
            actors = []
            for neurite in self.points[ntype]:
                # Get tree segments
                section_segment_list = [(section, segment)
                                                for section in iter_sections(neurite.component)
                                                for segment in iter_segments(section)]
                segs = [(seg[0][COLS.XYZ], seg[1][COLS.XYZ]) for _, seg in section_segment_list]

                # Get segments radius
                if fixed_neurite_radius:
                    if not isinstance(fixed_neurite_radius, (float , int)): 
                        raise ValueError(f"When passsing fixed_neurite_radius it should be a number not: {fixed_neurite_radius.__type__}")
                    radiuses = [self.points['soma'].radius * fixed_neurite_radius 
                                            for i in np.arange(len(segs))]
                else:
                    radiuses = _get_linewidth(neurite.component, diameter_scale=1, linewidth=1)
                    if not isinstance(radiuses, (tuple, list)): 
                        radiuses = [radiuses for i in np.arange(len(segs))]

                tubes = []
                for seg, rad in zip(segs, radiuses):
                    tubes.append(Tube(seg, rad))
                actors.append(merge([t.scale(1.05) for t in tubes]))
            neurites[ntype] = actors

        # Merge actors to get the entire neuron
        actors = [act.clone() for actors in neurites.values() for act in actors] + [soma.clone()]
        whole_neuron = merge(actors).clean().smoothMLS2D(f=0.1).c(whole_neuron_color)

        # Color actors
        [act.c(basal_dendrites_color) for act in neurites['basal_dendrites']]
        [act.c(apical_dendrites_color) for act in neurites['apical_dendrites']]
        [act.c(axon_color) for act in neurites['axon']]

        return neurites, whole_neuron



import os
import json
import requests
import yaml
import gzip
import numpy as np
import scipy.io as sio
import pandas as pd
from brainio import brainio
from vtkplotter import load, Volume

# ------------------------------------ OS ------------------------------------ #
def listdir(fld):
	"""
	List the files into a folder with the coplete file path instead of the relative file path like os.listdir.

	:param fld: string, folder path

	"""
	if not os.path.isdir(fld):
		raise FileNotFoundError("Could not find directory: {}".format(fld))

	return [os.path.join(fld, f) for f in os.listdir(fld)]



def get_subdirs(folderpath):
    """
		Returns the subfolders in a given folder
	"""
    return [f.path for f in os.scandir(folderpath) if f.is_dir()]


def check_file_exists(filepath, raise_error=False):
	# Check if a file with the given path exists already
	if os.path.isfile(filepath):
		return True
	elif raise_error:
		raise FileExistsError("File {} doesn't exist".format(filepath))
	else:
		return False

def get_file_name(filepath):
    # Returns just the name, no complete path or extension
    return os.path.splitext(os.path.basename(filepath))[0]

# ------------------------------ Load/Save data ------------------------------ #
def load_npy_from_gz(filepath):
	f = gzip.GzipFile(filepath, "r")
	return np.load(f)

def save_npy_to_gz(filepath, data):
	f = gzip.GzipFile(filepath, "w")
	np.save(f, data)
	f.close()

def save_json(filepath, content, append=False):
	"""
	Saves content to a JSON file

	:param filepath: path to a file (must include .json)
	:param content: dictionary of stuff to save

	"""
	if not 'json' in filepath:
		raise ValueError("filepath is invalid")

	if not append:
		with open(filepath, 'w') as json_file:
			json.dump(content, json_file, indent=4)
	else:
		with open(filepath, 'w+') as json_file:
			json.dump(content, json_file, indent=4)

def save_yaml(filepath, content, append=False, topcomment=None):
	"""
	Saves content to a yaml file

	:param filepath: path to a file (must include .yaml)
	:param content: dictionary of stuff to save

	"""
	if not 'yaml' in filepath:
		raise ValueError("filepath is invalid")

	if not append:
		method = 'w'
	else:
		method = 'w+'

	with open(filepath, method) as yaml_file:
		if topcomment is not None:
			yaml_file.write(topcomment)
		yaml.dump(content,yaml_file, default_flow_style=False, indent=4)

def load_json(filepath):
	"""
	Load a JSON file

	:param filepath: path to a file

	"""
	if not os.path.isfile(filepath) or not ".json" in filepath.lower(): 
		raise ValueError("unrecognized file path: {}".format(filepath))
	with open(filepath) as f:
		data = json.load(f)
	return data

def load_yaml(filepath):
	"""
	Load a YAML file

	:param filepath: path to yaml file

	"""
	if filepath is None or not os.path.isfile(filepath): raise ValueError("unrecognized file path: {}".format(filepath))
	if not "yml" in filepath and not "yaml" in filepath: raise ValueError("unrecognized file path: {}".format(filepath))
	return yaml.load(open(filepath), Loader=yaml.FullLoader)

def load_volume_file(filepath):
	"""
	Load a volume file (e.g., .nii) and returns the data

	:param filepath: path to file
	:param **kwargs: 

	"""
	if not os.path.isfile(filepath): raise FileNotFoundError(filepath)
	try:
		volume = brainio.load_any(filepath)
	except:
		raise ValueError(f"Could not load volume data: {filepath}")
	else:
		return volume

def load_mesh_from_file(filepath, *args, **kwargs):
	"""	
	Load a a mesh or volume from files like .obj, .stl, ...

	:param filepath: path to file
	:param **kwargs: 

	"""
	if not os.path.isfile(filepath): 
		raise FileNotFoundError(filepath)

	try:
		actor = load(filepath, *args, **kwargs)
	except:
		actor = Volume(load_volume_file(filepath, *args, **kwargs))
	
	return actor


# ----------------------------- Internet queries ----------------------------- #
def connected_to_internet(url='http://www.google.com/', timeout=5):
	"""
		Check that there is an internet connection

		:param url: url to use for testing (Default value = 'http://www.google.com/')
		:param timeout:  timeout to wait for [in seconds] (Default value = 5)
	"""
	
	try:
		_ = requests.get(url, timeout=timeout)
		return True
	except requests.ConnectionError:
		print("No internet connection available.")
	return False

def send_query(query_string, clean=False):
	"""
	Send a query/request to a website

	:param query_string: string with query content
	:param clean:  (Default value = False)

	"""
	response = requests.get(query_string)
	if response.ok:
		if not clean:
			return response.json()['msg']
		else:
			return response.json()
	else:
		raise ValueError("Invalide query string: {}".format(query_string))







# ---------------------------------------------------------------------------- #
#                               Data manipulation                              #
# ---------------------------------------------------------------------------- #
def flatten_list(lst):
    """
    Flattens a list of lists
    
    :param lst: list

    """
    flatten = []
    for item in lst:
        if isinstance(item, list):
            flatten.extend(item)
        else:
            flatten.append(item)
    return flatten

def is_any_item_in_list(L1, L2):
    """
    Checks if an item in a list is in another  list

    :param L1: 
    :param L2: 

    """
    # checks if any item of L1 is also in L2 and returns false otherwise
    inboth = [i for i in L1 if i in L2]
    if inboth:
        return True
    else:
        return False






from vtkplotter import Plotter

from morphapi.morphology.morphology import Neuron
from morphapi.api.neuromorphorg import NeuroMorpOrgAPI


api = NeuroMorpOrgAPI()

# ---------------------------- Downloading metadata --------------------------- #
# Get metadata for pyramidal neurons from the mouse cortex. 
metadata, _ = api.get_neurons_metadata(size=10,  # Can get the metadata for up to 500 neurons at the time
                    species='mouse', 
                    cell_type='pyramidal',
                    brain_region='neocortex')

# To get a list of available query fields: print(api.fields)
# To get a list of valid values for a field: print(api.get_fields_values(field))

print('Neurons metadata:')
print(metadata[0])

# ---------------------------- Download morphology --------------------------- #
neurons = api.download_neurons(metadata[5]) # Render only one neuron to speed things up. 


# ------------------------------- Visualisation ------------------------------ #
print('creating meshes')
neurons = [neuron.create_mesh()[1] for neuron in neurons]

print('visualizing')
vp = Plotter(shape=(1, len(neurons)), axes=1)

vp.show(neurons)


from vtkplotter import Plotter

from morphapi.morphology.morphology import Neuron

"""
    This example shows how to use vtkplotter to visualise a 3d reconstruction of a neuron.
    However, the reccomended way to visualise neurons is with brainrender:
    https://github.com/BrancoLab/BrainRender
"""

fp = 'examples/example_files/example1.swc'

# Create vtkplotter actors from the .swc file
neuron = Neuron(swc_file=fp)
components, neuron = neuron.create_mesh(fixed_neurite_radius=3, # Set this as None to use the actual radius
                soma_color='salmon', # Specify colors [see vtkplotter.colors for more details]
                apical_dendrites_color='darkseagreen',
                basal_dendrites_color='orangered',
                axon_color='blackboard',
                whole_neuron_color='blackboard')

# components stores an actor for each neuronal component (dendrites, soma...)
# neuron is a single actor for the entire neuron


# Show neuron as individual components or entire neuron
vp = Plotter(shape=(1, 2), axes=1)
vp.show(*list(components.values()), at=0, interactive=False)
vp.show(neuron, at=1,interactive=True)





import sys
sys.path.append('./')

import os
import pandas as pd
import numpy as np
from collections import namedtuple

from vtkplotter import shapes, load, merge

from allensdk.core.cell_types_cache import CellTypesCache
from allensdk.api.queries.cell_types_api import CellTypesApi
from allensdk.core.swc import Morphology

from morphapi.paths_manager import Paths
from morphapi.utils.data_io import connected_to_internet, listdir
from morphapi.morphology.morphology import Neuron


class AllenMorphology(Paths):
	""" Handles the download of neuronal morphology data from the Allen database. """

	def __init__(self, *args,  **kwargs):
		"""
			Initialise API interaction and fetch metadata of neurons in the Allen Database. 
		"""
		if not connected_to_internet():
			raise ConnectionError("You will need to be connected to the internet to use the AllenMorphology class to download neurons")

		Paths.__init__(self, *args, **kwargs)

		# Create a Cache for the Cell Types Cache API
		self.ctc = CellTypesCache(manifest_file=os.path.join(self.allen_morphology_cache, 'manifest.json'))

		# Get a list of cell metadata for neurons with reconstructions, download if necessary
		self.neurons = pd.DataFrame(self.ctc.get_cells(species=[CellTypesApi.MOUSE], require_reconstruction = True))
		self.n_neurons = len(self.neurons)

		if not self.n_neurons: 
			raise ValueError("Something went wrong and couldn't get neurons metadata from Allen")

		self.downloaded_neurons = self.get_downloaded_neurons()


	def get_downloaded_neurons(self):
		""" 
			Get's the path to files of downloaded neurons
		"""
		return [os.path.join(self.allen_morphology_cache, f) for f in os.listdir(self.allen_morphology_cache) if ".swc" in f]    


	def download_neurons(self, ids, **kwargs):
		"""
			Download neurons and return neuron reconstructions (instances
			of Neuron class)

		:param ids: list of integers with neurons IDs

		"""
		if isinstance(ids, np.ndarray):
			ids = list(ids)
		if not isinstance(ids, (list)): ids = [ids]

		neurons = []
		for neuron_id in ids:
			neuron_file = os.path.join(self.allen_morphology_cache, "{}.swc".format(neuron_id))

			# Download file 
			self.ctc.get_reconstruction(neuron_id, file_name=neuron_file)

			# Reconstruct neuron
			neurons.append(Neuron(neuron_file, **kwargs))
		
		return neurons



import sys
import os

from morphapi.utils.data_io import save_json

"""
    Class to create and store paths to a number of folders uesed to save/load data
"""


# Default paths for Data Folders (store stuff like object meshes, neurons morphology data etc)
default_paths = dict(
        # APIs caches
        allen_morphology_cache = 'Data/allen_morphology_cache',
        mouselight_cache = 'Data/mouselight_cache',
        neuromorphorg_cache = 'Data/neuromorphorg_cache',

        # Other
        mouse_connectivity_cache = 'Data/mouse_connectivity_cache'
)


class Paths:
    _folders = ["allen_morphology_cache", 
                "mouselight_cache", 
                "neuromorphorg_cache",
                'mouse_connectivity_cache', 
                ]

    def __init__(self, base_dir=None, **kwargs):
        """
        Parses a YAML file to get data folders paths. Stores paths to a number of folders used throughtout morphapi. 
        
        :param base_dir: str with path to directory to use to save data. If none the user's base directiry is used. 
        :param kwargs: use the name of a folder as key and a path as argument to specify the path of individual subfolders
        """
        # Get and make base directory
        if base_dir is None:
            user_dir = os.path.expanduser("~")
            if not os.path.isdir(user_dir):
                raise FileExistsError("Could not find user base folder (to save data). Platform: {}".format(sys.platform))
            self.base_dir = os.path.join(user_dir, ".morphapi")
        else:
            self.base_dir = base_dir

        if not os.path.isdir(self.base_dir):
            os.mkdir(self.base_dir)

        for fld_name in self._folders:
            # Check if user provided a path for this folder, otherwise use default
            fld_path = kwargs.pop(fld_name, default_paths[fld_name])

            # Make complete path and save it as an attribute of this class
            path = os.path.join(self.base_dir, fld_path)

            # Create folder if it doesn't exist 
            if not os.path.isdir(path):
                print("Creating folder at: {}".format(path))
                os.makedirs(path)
            self.__setattr__(fld_name, path)



import os

from morphapi.utils.webqueries import request, connected_to_internet
from morphapi.paths_manager import Paths
from morphapi.morphology.morphology import Neuron


class NeuroMorpOrgAPI(Paths):
    _base_url = 'http://neuromorpho.org/api/neuron'

    def __init__(self, *args, **kwargs):
        if not connected_to_internet():
            raise ConnectionError("You will need to be connected to the internet to use the NeuroMorpOrgAPI class to download neurons")

        Paths.__init__(self, *args, **kwargs)

        # Fields contains the types of fields that can be used to restrict queries
        self.fields = request(self._base_url+'/fields').json()['Neuron Fields']

    def get_fields_values(self, field):
        """
            Returns the list of allowed values for a given query field
        """
        return list(request(self._base_url+f'/fields/{field}').json()['fields'])

    def get_neurons_metadata(self, size=100, page=0,  **criteria):
        """
            Uses the neuromorpho API to download metadata about neurons.
            Criteri can be used to restrict the search to neurons of interest/
            http://neuromorpho.org/apiReference.html

            Neuromorpho.org  paginates it's requests so not all neurons metadata
            can be returned at once

            :param size: int in range [0, 500]. Number of neurons whose metadata can be returned at the same time
            :param page: int > 0. Page number. The number of pages depends on size and on how many neurons match the criteria
            :param criteria: use keywords to restrict the query to neurons that match given criteria.
                    keywords should be pass as "field=value". Then only neuron's whose 'field'
                    attribute has value 'value' will be returned.
        """

        if size < 0 or size>500:
            raise ValueError(f"Invalid size argument: {size}. Size should be an integer between 0 and 500")
        if page < 0:
            raise ValueError(f"Invalid page argument: {page}. Page should be an integer >= 0")


        url = self._base_url+'/select?q='

        for n, (crit, val) in enumerate(criteria.items()):
            if crit not in self.fields:
                raise ValueError(f"Query criteria {crit} not in available fields: {self.fields}")
            elif val not in self.get_fields_values(crit):
                raise ValueError(f"Query criteria value {val} for field {crit} not valid."+
                        f"Valid values include: {self.get_fields_values(crit)}")

            if isinstance(val, list):
                raise NotImplementedError('Need to join the list')

            if n>0:
                url += '&'
            url += f'{crit}:{val}'

        url += f'&size={int(size)}&page={int(page)}'
        req = request(url)

        if not req.ok:
            raise ValueError(f'Invalid query with url: {url}')
        else:
            neurons = req.json()

        page = neurons['page']    
        neurons = neurons['_embedded']['neuronResources']

        print(f"Found metadata for {page['totalElements']} neurons [{page['totalPages']} pages in total]"+
                f"\nReturning metadata about {len(neurons)} neurons from page {page['number']}")

        return neurons, page

    def get_neuron_by_id(self, nid):
        """
            Get a neuron's metadata given it's id number
        """
        return request(self._base_url+f'/id/{nid}').json()

    def get_neuron_by_name(self, nname):
        """
            Get a neuron's metadata given it's name
        """
        return request(self._base_url+f'/name/{nname}').json()

    def download_neurons(self, neurons):
        """
            Downloads neuronal morphological data and saves it to .swc files. 
            It then returns a list of Neuron instances with morphological data for each neuron.

            :param neurons: list of neurons metadata (as returned by one of the functions
                        used to fetch metadata)
        """
        if not isinstance(neurons, (list, tuple)):
            neurons  = [neurons]
        
        to_return = []
        for neuron in neurons:
            if not isinstance(neuron, dict):
                raise ValueError()

            filepath = os.path.join(self.neuromorphorg_cache, f"{neuron['neuron_id']}.swc")

            if not os.path.isfile(filepath):
                # Download and write to file
                url = f"http://neuromorpho.org/dableFiles/{neuron['archive'].lower()}/CNG version/{neuron['neuron_name']}.CNG.swc"

                req = request(url)
                if not req.ok:
                    raise ValueError(f"Failed to getch morphology data for neuron: {neuron['name']}")

                with open(filepath, 'w') as f:
                    f.write(req.content.decode('utf-8'))

            to_return.append(Neuron(filepath))
            
        return to_return







import json
import os 
from tqdm import tqdm
import pandas as pd
from collections import namedtuple
import numpy as np

from allensdk.core.mouse_connectivity_cache import MouseConnectivityCache

from morphapi.morphology.morphology import Neuron
from morphapi.utils.webqueries import post_mouselight, mouselight_base_url
from morphapi.paths_manager import Paths
from morphapi.utils.data_io import is_any_item_in_list
from morphapi.api.neuromorphorg import NeuroMorpOrgAPI


"""
    Collections of functions to query http://ml-neuronbrowser.janelia.org/ and get data about either the status of the API, 
    the brain regions or the neurons available. 
    Queries are sent by sending POST requests to http://ml-neuronbrowser.janelia.org/graphql
    with a string query. 
"""

# ---------------------------------------------------------------------------- #
#                                  QUERY UTILS                                 #
# ---------------------------------------------------------------------------- #

def mouselight_api_info():
    """
        Get the number of cells available in the database
    """
    # Get info from the ML API
    url = mouselight_base_url + "graphql"

    query =  \
            """
            query {
                queryData {
                    totalCount
                    }
                }
            """
    res =  post_mouselight(url, query=query)
    print("{} neurons on MouseLight database. ".format(res['queryData']['totalCount']))

def mouselight_get_brainregions():
    """
        Get metadata about the brain brain regions as they are known by Janelia's Mouse Light. 
        IDs and Names sometimes differ from Allen's CCF.
    """

    # Download metadata about brain regions from the ML API
    url = mouselight_base_url + "graphql"
    # query =  "systemSettings {apiVersion apiRelease neuronCount\}}"
    query =  \
            """
            query {
                brainAreas{
                    
                    acronym
                    name
                    id
                    atlasId
                    graphOrder
                    parentStructureId
                    structureIdPath
                }
            }
            """
    res =  post_mouselight(url, query=query)['brainAreas']

    # Clean up and turn into a dataframe
    keys = {k:[] for k in res[0].keys()}
    for r in res:
        for k in r.keys():
            keys[k].append(r[k])
    
    structures_data = pd.DataFrame.from_dict(keys)
    return structures_data

def mouselight_structures_identifiers():
    """
    When the data are downloaded as SWC, each node has a structure identifier ID to tell if it's soma, axon or dendrite.
    This function returns the ID number --> structure table. 
    """

    # Download the identifiers used in ML neurons tracers
    url = mouselight_base_url + "graphql"
    # query =  "systemSettings {apiVersion apiRelease neuronCount\}}"
    query =  \
        """
            query {
                structureIdentifiers{
                    id
                    name
                    value
                }
            }
        """
    res =  post_mouselight(url, query=query)['structureIdentifiers']

    keys = {k:[] for k in res[0].keys()}
    for r in res:
        for k in r.keys():
            keys[k].append(r[k])
    
    structures_identifiers = pd.DataFrame.from_dict(keys)
    return structures_identifiers

def make_query(filterby=None, filter_regions=None, invert=False):
    """
    Constructs the strings used to submit graphql queries to the mouse light api

    :param filterby: str, soma, axon on dendrite. Search by neurite structure (Default value = None)
    :param filter_regions:  list, tuple. list of strings. Acronyms of brain regions to use for query (Default value = None)
    :param invert:  If true the inverse of the query is return (i.e. the neurons NOT in a brain region) (Default value = False)

    """
    searchneurons = """
                queryTime
                totalCount
                
                neurons{
                tag
                id
                idNumber
                idString
                
                brainArea{
                    id
                    acronym
                    name
                    safeName
                    atlasId
                    aliases
                    structureIdPath
                }

                tracings{
                    soma{
                    x
                    y
                    z
                    radius
                    brainArea{
                        id
                        acronym
                    }
                    sampleNumber
                    parentNumber
                    
                    }
                
                id
                tracingStructure{
                    name
                    value
                    id
                }
                }
            }
    """


    if filterby is None or filterby == 'soma':
        query = """
                    query {{
                        searchNeurons {{
                            {}
                        }}
                    }}
                    """.format(searchneurons)
    else:
        raise NotImplementedError("This feature is not available yet")
        # Get predicate type
        if filterby.lower() in ['axon','axons', 'end point', 'branch point']:     
            predicateType = 1
            
        elif filterby.lower() in ['dendrite', 'apical dendrite', '(basal) dendrite']:
            raise NotImplementedError
            filterby = "(basal) dendrite"
            predicateType = 2
        else:
            raise ValueError("invalid search by argument")

        # Get neuron structure id
        structures_identifiers = mouselight_structures_identifiers()
        structureid = str(structures_identifiers.loc[structures_identifiers.name == filterby]['id'].values[0])

        # Get brain regions ids
        brainregions = mouselight_get_brainregions()
        brainareaids = [str(brainregions.loc[brainregions.acronym == a]['id'].values[0]) for a in filter_regions]

        # Get inversion
        if invert:
            invert = "true"
        else:
            invert = "false"

        query = """
        query {{
            searchNeurons (
                context: {{
                scope: 6
                predicates: [{{
                    predicateType: {predicate}
                    tracingIdsOrDOIs: []
                    tracingIdsOrDOIsExactMatch: false
                    tracingStructureIds: []
                    amount: 0
                    nodeStructureIds: ['{structure}']
                    brainAreaIds: {brainarea}
                    invert: {invert}
                    composition: 1
                    }}]
                }}
            ) {{
                {base}
            }}
        }}
        """.format(predicate=predicateType,  structure=str(structureid), brainarea=brainareaids, invert=invert, base=searchneurons)

        query = query.replace("\\t", "").replace("'", '"')
    return query
    



# ---------------------------------------------------------------------------- #
#                                  MAIN CLASS                                  #
# ---------------------------------------------------------------------------- #

class MouseLightAPI(Paths):
    def __init__(self, base_dir=None, **kwargs):
        """
            Handles the download of neurons morphology data from the Mouse Light project

            :param base_dir: path to directory to use for saving data (default value None)
            :param kwargs: can be used to pass path to individual data folders. See morphapi/utils /paths_manager.py
        """
        Paths.__init__(self, base_dir=base_dir, **kwargs)


    def fetch_neurons_metadata(self, filterby = None, filter_regions=None, **kwargs):
        """
        Download neurons metadata and data from the API. The downloaded metadata can be filtered to keep only
        the neurons whose soma is in a list of user selected brain regions.
        
        :param filterby: Accepted values: "soma". If it's "soma", neurons are kept only when their soma
                        is in the list of brain regions defined by filter_regions (Default value = None)
        :param filter_regions: List of brain regions acronyms. If filtering neurons, these specify the filter criteria. (Default value = None)
        :param **kwargs: 

        """
        # Download all metadata
        print("Querying MouseLight API...")
        url = mouselight_base_url + "graphql"
        query = make_query(filterby=filterby, filter_regions=filter_regions, **kwargs)

        res =  post_mouselight(url, query=query)['searchNeurons']
        print("     ... fetched metadata for {} neurons in {}s".format(res["totalCount"], round(res["queryTime"]/1000, 2)))

        # Process neurons to clean up the results and make them easier to handle
        neurons = res['neurons']
        node = namedtuple("node", "x y z r area_acronym sample_n parent_n")
        tracing_structure = namedtuple("tracing_structure", "id name value named_id")

        cleaned_nurons = [] # <- output is stored here
        for neuron in neurons:
            if neuron['brainArea'] is not None:
                brainArea_acronym = neuron['brainArea']['acronym']
                brainArea_id = neuron['brainArea']['id']
                brainArea_name = neuron['brainArea']['name']
                brainArea_safename = neuron['brainArea']['safeName']
                brainArea_atlasId = neuron['brainArea']['atlasId']
                brainArea_aliases = neuron['brainArea']['aliases']
                brainArea_structureIdPath = neuron['brainArea']['structureIdPath']
            else:
                brainArea_acronym = None
                brainArea_id = None
                brainArea_name = None
                brainArea_safename = None
                brainArea_atlasId = None
                brainArea_aliases = None
                brainArea_structureIdPath = None

            if len(neuron['tracings']) > 1:
                dendrite = tracing_structure(
                    neuron['tracings'][1]['id'],
                    neuron['tracings'][1]['tracingStructure']['name'],
                    neuron['tracings'][1]['tracingStructure']['value'],
                    neuron['tracings'][1]['tracingStructure']['id'],
                )
            else:
                dendrite = None

            clean_neuron = dict(
                brainArea_acronym = brainArea_acronym,
                brainArea_id = brainArea_id,
                brainArea_name = brainArea_name,
                brainArea_safename = brainArea_safename,
                brainArea_atlasId = brainArea_atlasId,
                brainArea_aliases = brainArea_aliases,
                brainArea_structureIdPath = brainArea_structureIdPath,

                id = neuron['id'],
                idNumber = neuron['idNumber'],
                idString = neuron['idString'],
                tag=neuron['tag'],
                soma = node(
                    neuron['tracings'][0]['soma']['x'],
                    neuron['tracings'][0]['soma']['y'],
                    neuron['tracings'][0]['soma']['z'],
                    neuron['tracings'][0]['soma']['radius'],
                    neuron['tracings'][0]['soma']['brainArea'],
                    neuron['tracings'][0]['soma']['sampleNumber'],
                    neuron['tracings'][0]['soma']['parentNumber']
                ),
                axon = tracing_structure(
                    neuron['tracings'][0]['id'],
                    neuron['tracings'][0]['tracingStructure']['name'],
                    neuron['tracings'][0]['tracingStructure']['value'],
                    neuron['tracings'][0]['tracingStructure']['id'],
                ),
                dendrite = dendrite,
            )
            cleaned_nurons.append(clean_neuron)

        # Filter neurons to keep only those matching the search criteria
        if filterby is not None:
            if filter_regions is None:
                raise ValueError("If filtering neuron by region, you need to pass a list of filter regions to use")

            # Get structure tree 
            smc = MouseConnectivityCache(manifest_file=os.path.join(self.mouse_connectivity_cache, "manifest.json"))
            structure_tree = smc.get_structure_tree()
            ancestors_tree = structure_tree.get_ancestor_id_map()
            filter_regions_ids = [struct['id'] for struct in structure_tree.get_structures_by_acronym(filter_regions)]

            # Filter by soma
            if filterby == "soma":
                filtered_neurons = []
                for neuron in cleaned_nurons:
                    if neuron['brainArea_acronym'] is None: 
                        continue

                    # Get region ID (of the soma) and the IDs of its ancestors
                    region = structure_tree.get_structures_by_acronym([neuron['brainArea_acronym']])[0]
                    region_ancestors = ancestors_tree[region['id']]

                    # If any of the ancestors are in the allowed regions, keep neuron.
                    if is_any_item_in_list(filter_regions_ids, region_ancestors):
                        filtered_neurons.append(neuron)
                print("	... selected {} neurons out of {}".format(len(filtered_neurons), res["totalCount"]))
                
                neurons =  filtered_neurons
            else:
                print("	... selected {} neurons out of {}".format(len(cleaned_nurons), res["totalCount"]))
                neurons =  cleaned_nurons
        else:
            neurons = cleaned_nurons

        return neurons


    def download_neurons(self, neurons_metadata):
        """
        Given a list of neurons metadata from self.fetch_neurons_metadata
        this funcition downloads the morphological data.
        The data are actually downloaded from neuromorpho.org
        
        :param neurons_metadata: list with metadata for neurons to download
        :returns: list of Neuron instances

        """
        if not isinstance(neurons_metadata, (list, tuple)):
            neurons_metadata = [neurons_metadata]
        nmapi = NeuroMorpOrgAPI()

        neurons = []
        for neuron in tqdm(neurons_metadata):
            neurons.append(nmapi.download_neurons(nmapi.get_neuron_by_name(neuron['idString'])))

        # ? Old code
        # def get(url, tracing_id): # send a query for a single tracing ID
        #     """
        #     Creates the URL for each neuron to download

        #     :param url: str with url
        #     :param tracing_id: str with the neuron's ID

        #     """
        #     query = {"ids":[tracing_id]}
        #     res = post_mouselight(url, query=query, clean=True)['tracings']
        #     return res

        # if neurons_metadata is None or not neurons_metadata:
        #     return None
        # print("Downloading neurons tracing data from Mouse Light")

        # # check arguments
        # if not isinstance(neurons_metadata, list): neurons_metadata = [neurons_metadata]

        # # URL used to fetch neurons
        # url = mouselight_base_url + "tracings/tracings"

        # # loop over neurons
        # neurons = []
        # for neuron in tqdm(neurons_metadata):
        #     # Check if a .swc file already exists for this neuron
        #     file_path = os.path.join(self.mouselight_cache, neuron['idString']+".swc")
            
        #     if not os.path.isfile(file_path): 
        #         # Download and save as swc
        #         # Get tracings by requests
        #         axon_tracing, dendrite_tracing = None, None
        #         if neuron['axon'] is not None:
        #             axon_tracing = get(url, neuron['axon'].id)
        #         if neuron['dendrite'] is not None:
        #             dendrite_tracing = get(url, neuron['dendrite'].id)

        #         self.make_swc(neuron, file_path, axon_tracing=axon_tracing, dendrite_tracing=dendrite_tracing)

        #     # Reconstruct from .swc
        #     neurons.append(Neuron(file_path,))

        return neurons







"""
Old code to fetch morphological data from mouselight database
"""
    # @staticmethod
    # def make_json(neuron, file_path,  axon_tracing=None, dendrite_tracing=None):
    #     """
    #     Creates a .json file with the neuron's data that can be read by the mouselight parser.

    #     :param neuron: dict with neuron's data
    #     :param file_path: str, path where to save the json file
    #     :param axon_tracing: list with data for axon tracing (Default value = None)
    #     :param dendrite_tracing: list with data for dendrite tracing (Default value = None)

    #     """
    #     # parse axon
    #     if axon_tracing is not None:
    #         nodes = axon_tracing[0]['nodes']
    #         axon = [
    #             dict(
    #                 sampleNumber = n['sampleNumber'],
    #                 x = n['x'],
    #                 y = n['y'],
    #                 z = n['z'],
    #                 radius = n['radius'],
    #                 parentNumber = n['parentNumber'],
    #             )
    #             for n in nodes]
    #     else:
    #         axon = []

    #     # parse dendrites
    #     if dendrite_tracing is not None:
    #         nodes = dendrite_tracing[0]['nodes']
    #         dendrite = [
    #             dict(
    #                 sampleNumber = n['sampleNumber'],
    #                 x = n['x'],
    #                 y = n['y'],
    #                 z = n['z'],
    #                 radius = n['radius'],
    #                 parentNumber = n['parentNumber'],
    #             )
    #             for n in nodes]
    #     else:
    #         dendrite = []

    #     content = dict(
    #         neurons = [
    #             dict(
    #                 idString = neuron['idString'],
    #                 soma = dict(
    #                     x = neuron['soma'].x,
    #                     y = neuron['soma'].y,
    #                     z = neuron['soma'].z,
    #                 ),
    #                 axon = axon, 
    #                 dendrite = dendrite,

    #             )
    #         ]
    #     )


    #     # save to file
    #     if not filepath.endswith('.json'):
    #         raise ValueError(f'Invalid file path {filepath}')
    #     with open(file_path, 'w') as f:
    #         json.dump(content, f)

    # @staticmethod
    # def make_swc(neuron, file_path,  axon_tracing=None, dendrite_tracing=None):
    #     """
    #     Creates a .swc file with the neuron's data that can be read by the mouselight parser.

    #     :param neuron: dict with neuron's data
    #     :param file_path: str, path where to save the swc file
    #     :param axon_tracing: list with data for axon tracing (Default value = None)
    #     :param dendrite_tracing: list with data for dendrite tracing (Default value = None)

    #     """

    #     if not file_path.endswith('.swc'):
    #         raise ValueError(f'Invalid file path {file_path}')

    #     # parse data
    #     if axon_tracing is not None:
    #         nodes = axon_tracing[0]['nodes']
    #         axon = [
    #             dict(
    #                 sampleNumber = n['sampleNumber'],
    #                 x = n['x'],
    #                 y = n['y'],
    #                 z = n['z'],
    #                 id = 2,
    #                 radius = n['radius'],
    #                 parentNumber = n['parentNumber'],
    #             )
    #             for n in nodes]
    #     else:
    #         axon = []

    #     if dendrite_tracing is not None:
    #         nodes = dendrite_tracing[0]['nodes']
    #         dendrite = [
    #             dict(
    #                 sampleNumber = n['sampleNumber'],
    #                 x = n['x'],
    #                 y = n['y'],
    #                 z = n['z'],
    #                 radius = n['radius'],
    #                 id = 3,
    #                 parentNumber = n['parentNumber'],
    #             )
    #             for n in nodes]
    #     else:
    #         dendrite = []

    #     # Save as swc
    #     with open(file_path, 'w') as out:
    #         out.write(f"# neuron: {neuron['idString']} downloaded from mouselight database [janelia campus]\n")
    #         out.write("# .swc reconstructed using morphapi python package https://github.com/brainglobe/morphapi\n\n")
    #         out.write("# id,type,x,y,z,r,pid\n")
    #         out.write(f" 1 1 {neuron['soma'].x} "+
    #                     f"{neuron['soma'].y} {neuron['soma'].z} "+
    #                     f"{neuron['soma'].r} -1\n")


    #         neurites = axon + dendrite
    #         sort_idx = np.argsort([n['parentNumber'] for n in neurites])
    #         neurites = list(np.array(neurites)[sort_idx])

    #         for n in neurites:
    #             if n['parentNumber'] == -1:
    #                 pn = 1
    #             else:
    #                 pn = n['parentNumber']
    #             out.write(f" {n['sampleNumber']} {n['id']} {n['x']} {n['y']} {n['z']} {n['radius']} {pn}\n")



from vtkplotter import Plotter

from morphapi.morphology.morphology import Neuron
from morphapi.api.mouselight import MouseLightAPI, mouselight_structures_identifiers



# ---------------------------- Downloading neurons --------------------------- #
mlapi = MouseLightAPI()

# Fetch metadata for neurons with some in the secondary motor cortex
neurons_metadata = mlapi.fetch_neurons_metadata(filterby='soma', filter_regions=['MOs'])

# Then we can download the files and save them as a .json file
neurons_files =  mlapi.download_neurons(neurons_metadata[0]) # just saving the first couple neurons to speed things up



"""
    mlapi.download_neurons returns a list of instances of the class Neuron
    from morphapi.morphology.morphology.
"""


# ------------------------------- Visualisation ------------------------------ #
print('creating meshes')
neurons = [neuron.create_mesh()[1] for neuron in neurons]

print('visualizing')
vp = Plotter(shape=(1, len(neurons)), axes=1)

vp.show(neurons)